import os

class BaseCIAdapter():

    def __init__(self, logger=None):
        self.logger = logger
        self.env_map = {}

    def convert_environment_variables(self):
        for key, value in os.environ.items():
            if key in self.env_map and not self.env_map[key] in os.environ:
                os.environ[self.env_map[key]] = value
                os.environ.pop(key)

    def write_file(self, path, text):
        file_text = """
/*************************************************
 * DO NOT EDIT THIS FILE, see .servicer/ instead *
 * CI Configuration Auto-generated by Servicer   *
 *************************************************/

"""
        file_text += text

        with open(path, "w") as text_file:
            text_file.write(file_text)

        self.logger.log('CI config file generated: %s' % path)



import sys
import os
from urllib.parse import quote_plus

from .run import run

class Git():
    def __init__(self, hide_output=True, protocol='ssh', logger=None):
        self.run = run
        self.logger = logger
        self.hide_output = hide_output
        self.protocol = os.getenv('GIT_PROTOCOL', protocol)

    def authors_for_changes_ahead_of_ref(self, ref):
        forward_commits = self.commits_against_ref(ref, commit_type='+')
        authors = set()
        for c in forward_commits:
            authors.add(self.author_for_ref(c))
        return sorted(list(authors))

    def commits_against_ref(self, ref, commit_type='', hide_output=None):
        result = self.run('git cherry %s' % ref, hide_output=hide_output or self.hide_output)

        commits = result['stdout'].strip().split('\n')
        return [c[2:] for c in commits if c and c.startswith(commit_type)]

    def diff(self, a, b='HEAD', name_only=False, merge_base=False, hide_output=None):
        command = 'git diff %s' % a
        if merge_base:
            command = '%s...%s' % (command, b)
        else:
            command = '%s..%s' % (command, b)
        if name_only:
            command = '%s --name-only' % command

        result = self.run(command, hide_output=hide_output or self.hide_output)

        if name_only:
            return [f for f in result['stdout'].strip().split('\n') if len(f) > 0]

        return result['stdout'].strip()

    def tag(self, tag, message=None, push=False, hide_output=None):
        command = 'git tag'
        if message:
            command = '%s -a %s -m "%s"' % (command, tag, message)
        else:
            command = '%s %s' % (command, tag)

        self.run(command, hide_output=hide_output or self.hide_output)

        if push:
            self.push(ref=tag)

    def delete_tag(self, tags):
        if not isinstance(tags, list):
            tags = [tags]

        for tag in tags:
            self.run('git tag -d %s' % tag, check=False)
            self.push(ref=':refs/tags/%s' % tag)

    def list_tags(self):
        result = self.run('git tag', hide_output=True)
        return result['stdout'].strip().split('\n')

    def list_remote_branches(self):
        result = self.run('git branch -r', hide_output=True)
        return result['stdout'].strip().split('\n')

    def sanitize_tag(self, tag):
        for ch in '[ ]'.split():
            tag = tag.replace(ch, '.')
        return tag

    def set_config(self, config=None):
        for key, value in config.items():
            result = self.run('git config %s' % key)
            if not result['stdout'].strip():
                self.run('git config %s "%s"' % (key, value))

    def commit(self, add='.', message=None, check_commit=False):
        if not isinstance(add, list):
            add = [add]

        for a in add:
            self.run('git add %s' % a)

        return self.run('git commit -m "%s"' % message, check=check_commit)

    def push(self, origin='origin', ref=None, local_ref=None, protocol=None, no_verify=False):
        if protocol == None:
            protocol = self.protocol

        command = 'git push'

        if local_ref:
            ref = '%s:%s' % (ref, local_ref)

        if no_verify or ('no-verify' in self.config and self.config['no-verify']):
            command = '%s --no-verify' % command

        if protocol == 'ssh':
            command = '%s %s %s' % (command, origin, ref)
        elif protocol == 'https':
            command = '%s https://%s:%s@%s %s' % (
                command,
                os.environ['GIT_USERNAME'],
                quote_plus(os.environ['GIT_PASSWORD']),
                os.environ['GIT_REPOSITORY'],
                ref,
            )
        else:
            raise ValueError('Invalid git push protocol: %s' % protocol)

        return self.run(command)

    def current_branch(self, ref='HEAD'):
        command = 'git rev-parse --abbrev-ref %s' % ref
        result = self.run(command)
        return result['stdout'].strip()

    def current_commit(self, min_length=None, verify=False, ref='HEAD'):
        command = 'git rev-parse'
        if min_length:
            command = '%s --short=%s' % (command, min_length)
        if verify:
            command = '%s --verify' % command
        command = '%s %s' % (command, ref)

        result = self.run(command)
        return result['stdout'].strip()

    def author_for_ref(self, ref=None):
        command = 'git show --quiet --format="%an"'
        if ref:
            command = '%s %s' % (command, ref)

        result = self.run(command)
        return result['stdout'].strip()

    def fetch(self):
        command = 'git fetch'
        self.run(command)



import subprocess

def run(command, check=True, shell=True, hide_output=False):
    print('executing: %s' % command)
    result = { 'command': command }

    try:
        command_result = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=check,
            shell=shell,
        )
        # command_result = subprocess.run(command, check = check, shell = shell)
        result['command_result'] = command_result
        result['stdout'] = command_result.stdout.decode('utf-8')
        result['status'] = command_result.returncode

        if not hide_output:
            print(result['stdout'])

    except subprocess.CalledProcessError as e:
        print('failed: %s' % e.returncode)
        print(e.output.decode('utf-8'))
        raise

    return result



from .aws_service import AWSService
import time
import json

class Service(AWSService):
    def up(self):
        super().up()

        self.rds = self.boto3.client('rds')

        self.create_db_instance(self.config['rds_params'])
        database_host = self.wait_for_db_available(self.config['rds_params']['DBInstanceIdentifier'])

        return {
            'database_host': database_host,
        }

    def create_db_instance(self, params):
        from botocore.client import ClientError

        try:
            self.rds.create_db_instance(**params)
            self.logger.log('creating RDS instance with ID: %s' % params['DBInstanceIdentifier'])
        except ClientError as e:
            if 'DBInstanceAlreadyExists' in str(e):
                self.logger.log('DB instance (%s) exists already, continuing to poll ...' % id)
            else:
                raise

    def wait_for_db_available(self, id, attempts=120, sleep_time=10):
        instance = self.get_instance(id)
        while instance['DBInstanceStatus'] != 'available' and attempts > 0:
            time.sleep(sleep_time)
            instance = self.get_instance(id)
            attempts -= 1
            self.logger.log('status: %s, remaining attempts: %s' % (instance['DBInstanceStatus'], attempts))

        self.logger.log('DB instance (%s) ready at: %s' % (id, instance['Endpoint']['Address']))
        return instance['Endpoint']['Address']

    def get_instance(self, id):
        response = self.rds.describe_db_instances(DBInstanceIdentifier=id)
        return response['DBInstances'][0]

    # def down():
    #     print('TODO!')



import sys
import requests
import json

from ..task_service import Service as BaseService
from ..request_service import Service as RequestService

class Service(BaseService, RequestService):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)

        self.base_url = self.config.get('endpoint', os.environ['VAULT_ENDPOINT'])
        self.token = self.config.get('token', os.environ['VAULT_TOKEN'])
        self.base_path = self.config.get('base_path', '')

    def ensure_secret_engine(self, **params):
        response = self.vault_request(url='sys/mounts')
        engines = response.json().keys()
        self.logger.log('engines: %s' % engines)

        if '%s/' % params['path'] not in engines:
            self.logger.log('creating secrets engine: %s(%s)' % (params['path'], params['type']))
            self.logger.log(params)
            response = self.vault_request(
                method='post',
                url='sys/mounts/%s' % params['path'],
                json=params,
            )
            self.logger.log(response.text)
        else:
            self.logger.log('secrets engine already exists: %s(%s)' % (params['path'], params['type']))

    def ensure_policy(self, **params):
        name = params.pop('name')
        self.logger.log('creating policy %s' % name)
        self.logger.log(params)
        response = self.vault_request(
            method='post',
            url='sys/policy/%s' % name,
            json=params,
        )
        self.logger.log(response.text)

    def write(self, **params):
        path = params.pop('path')
        response = self.vault_request(
            method='post',
            url='%s%s' % (self.base_path, path),
            json=params,
        )
        self.logger.log(response.text)

    def read(self, **params):
        response = self.vault_request(url='%s%s' % (self.base_path, params['path']))
        self.logger.log(response.text)

    def vault_request(self, **params):
        if 'headers' not in params:
            params['headers'] = {}
        params['headers']['X-Vault-Token'] = self.token

        return self.request(**params)



from .gcloud_service import GCloudService
import os
import requests
from google.cloud import storage

class Service(GCloudService):
    def up(self):
        super().up()

        self.project = self.config.get('project', os.environ['GCLOUD_PROJECT'])
        self.gs = storage.Client(project=self.project)

        self.bucket = self.config['bucket']
        self.ensure_bucket()

        if 'steps' in self.config:
            self.run_steps(self.config['steps'])

    def down(self):
        super().down()

    def ensure_bucket(self):
        bucket_args = {
            'Bucket': self.config['bucket'],
        }

        self.logger.log('ensuring bucket exists: %s' % self.bucket)
        bucket = storage.Bucket(self.gs, name=self.bucket)

        region = self.config.get('region', os.getenv('GCLOUD_REGION'))
        if region:
            bucket.location = region

        if not bucket.exists():
            bucket.create()

    def run_steps(self, steps):
        for step in steps:
            getattr(self, step['type'])(**step['args'])



import os
import json
import yaml

from .token_interpolator import TokenInterpolator

class ConfigLoader():
    def __init__(self, args={}, logger=None):
        self.args = args
        self.token_interpolator = TokenInterpolator()
        self.logger = logger
        self.module_path = os.path.dirname(os.path.realpath(globals()['__file__']))
        self.servicer_config_path = args.get('servicer_config_path')
        self.servicer_config_file_path = '%s/%s' % (self.servicer_config_path, args.get('services_file'))

    def load_config(self):
        services_config = {}

        if self.servicer_config_path and self.servicer_config_file_path:
            self.logger.log('loading services config from (%s)' % self.servicer_config_file_path)
            self.load_extended_config(config_path=self.servicer_config_file_path, config=services_config)
            services_config = self.merge_defaults(config=services_config)
            self.merge_included_configs(config_path=self.servicer_config_file_path, config=services_config)

            services_config['config_path'] = self.servicer_config_path

        services_config['module_path'] = self.module_path
        services_config['args'] = self.args

        if 'environment' in services_config and 'variables' in services_config['environment']:
            self.load_environment_variables(services_config['environment']['variables'])

        return services_config

    # recursively load configs, overwriting base config values
    def load_extended_config(self, config_path=None, config=None):
        merge_config = yaml.load(open(config_path))

        if 'extends' in merge_config:
            config_path_pieces = config_path.split('/')
            inherit_path = '%s/%s' % ('/'.join(config_path_pieces[:-1]), merge_config.pop('extends'))
            self.logger.log('Extending: %s' % inherit_path)
            self.load_extended_config(config_path=inherit_path, config=config)

        self.merge_config(config, merge_config)

    def merge_included_configs(self, config_path=None, config=None):
        iterable = None

        if isinstance(config, dict):
            self.load_include_configs(config_path=config_path, config=config)
            iterable = config.values()
        elif isinstance(config, list):
            iterable = config

        if iterable:
            for value in iterable:
                self.merge_included_configs(config_path=config_path, config=value)

    def load_include_configs(self, config_path=None, config=None):
        if 'includes' not in config:
            return

        includes = config.pop('includes')
        if not isinstance(includes, list):
            includes = [includes]

        for include in includes:
            path = include
            params = {}

            if isinstance(include, dict):
                path = include['path']
                params = include['params']

            config_path_pieces = config_path.split('/')
            include_path = '%s/%s' % ('/'.join(config_path_pieces[:-1]), path)

            self.logger.log('Including: %s' % include_path)

            include_config = {}
            self.load_extended_config(config_path=include_path, config=include_config)
            self.merge_included_configs(config_path=self.servicer_config_file_path, config=include_config)

            if params:
                self.token_interpolator.interpolate_tokens(include_config, params, ignore_missing_key=True)

            self.merge_config(config, include_config)

    def merge_defaults(self, config={}):
        default_config_path = '%s/builtin/defaults.yaml' % self.module_path
        default_config = yaml.load(open(default_config_path))
        self.merge_config(default_config, config)

        return default_config

    def merge_config(self, merge_to, merge_from):
        for key, value in merge_from.items():
            if isinstance(value, dict):
                # get node or create one
                node = merge_to.setdefault(key, {})
                self.merge_config(node, value)
            else:
                merge_to[key] = value

        return merge_to

    def interpolate_config(self, config):
        self.logger.log('Interpolating Tokens...')
        self.token_interpolator.interpolate_tokens(config, os.environ, ignore_missing_key=True)

    def load_environment_variables(self, variables={}):
        for key, value in variables.items():
            os.environ[key] = self.token_interpolator.replace_tokens(value, os.environ)



import re

class TokenInterpolator():
    def __init__(self, logger=None):
        self.logger = logger

    def interpolate_tokens(self, config, params, ignore_missing_key=False):
        if isinstance(config, dict):
            for key, value in config.items():
                if isinstance(value, str):
                    config[key] = self.replace_tokens(value, params, ignore_missing_key)
                else:
                    self.interpolate_tokens(value, params, ignore_missing_key)
        elif isinstance(config, list):
            for i in range(len(config)):
                if isinstance(config[i], str):
                    config[i] = self.replace_tokens(config[i], params, ignore_missing_key)
                else:
                    self.interpolate_tokens(config[i], params, ignore_missing_key)

    def replace_tokens(self, value, params, ignore_missing_key=False):
        escaped_values = list('.^$*+?()[]{}|')

        for match in re.findall(r'\${.+?}+', value):
            token = match[2:-1]

            replace_value = self.evaluate_token(match, params)

            if replace_value:
                for ev in escaped_values:
                    if ev in token:
                        token = token.replace(ev, '\%s' % ev)
                value = re.sub(r'\${%s}' % token, replace_value, value)

        return value

    def evaluate_token(self, value, params):
        if not (value.startswith('${') and value.endswith('}')):
            return value

        pieces = value[2:-1].split(':')
        if len(pieces) > 2:
            pieces[1] = ':'.join(pieces[1:])
            pieces = pieces[:2]

        for p in pieces:
            evaluated = self.evaluate_value(p, params)
            if evaluated:
                return evaluated

        return None

    def evaluate_value(self, value, params):
        if value.startswith('"') and value.endswith('"'):
            return value[1:-1]

        result = None
        path_pieces = value.split('.')
        if len(path_pieces) > 0:
            path_value = self.dict_get_path(_dict=params, path=path_pieces, ignore_missing_key=True)
            if path_value:
                result = path_value

        return result

    def dict_get_path(self, _dict={}, path=[], ignore_missing_key=False):
        if path == None or len(path) < 1:
            return None

        _path = path.copy()
        current_path = _path.pop(0)
        if ignore_missing_key and current_path not in _dict:
            return None

        if len(_path) < 1:
            return _dict[current_path]

        return self.dict_get_path(_dict=_dict[current_path], path=_path, ignore_missing_key=ignore_missing_key)



import sys
from .topological_order import toposort2

def generate_ci_config(config=None, path=None):
    if not config:
        print('You must provide a valid deploy config!')

    create_ci_steps(config)
    output_configs = config['ci']['output_configs']
    generators = [getattr(sys.modules[__name__], 'generate_%s' % oc['provider']) for oc in output_configs]
    results = [g(config, path) for g in generators]

def create_ci_steps(config):
    steps = []

    if 'commands' in config['ci']:
        steps.append(create_step(
            name='ci_setup',
            commands=config['ci']['commands'],
        ))

    build_and_test_steps = []
    for service in config['services'].values():
        print(service)
        commands = []
        for step in ('build', 'test'):
            if 'steps' in service and step in service['steps'] and 'commands' in service['steps'][step]:
                commands.extend(service['steps'][step]['commands'])
            elif service.get('module') and hasattr(service['module'], step):
                commands.append('python %s/deploy.py --step %s --service %s' % (config['path'], step, service['name']))
        if len(commands) > 0:
            build_and_test_steps.append(create_step(
                name='%s_build_and_test' % service['name'],
                commands=commands,
                docker=service.get('docker')
            ))

    if len(build_and_test_steps) > 0:
        steps.append(create_step(
            name='build_and_test',
            type='parallel',
            steps=build_and_test_steps,
        ))

    deploy_dependencies = parse_dependencies(config['services'])
    deploy_order = toposort2(deploy_dependencies)

    for parallel_step in deploy_order:
        deploy_steps = []
        for deploy_step in parallel_step:
            deploy_steps.append(create_step(
                name='%s_deploy' % deploy_step,
                commands=['python %s/deploy.py --step deploy --service %s' % (config['path'], deploy_step)],
                docker=config['services'][deploy_step].get('docker')
            ))

        steps.append(create_step(
            name='deploy_%s' % '_'.join(parallel_step),
            type='parallel',
            steps=deploy_steps,
        ))

    config['ci']['steps'] = steps
    return steps

def create_step(**args):
    step = {
        'name': args['name'],
        'type': args.get('type') or 'step',
    }
    for arg in ('commands', 'docker', 'steps'):
        if arg in args:
            step[arg] = args[arg]

    return step

def parse_dependencies(services):
    dependencies = {}

    for service in services.values():
        deploy_deps = service.get('depends_on', [])
        if not isinstance(deploy_deps, list):
            deploy_deps = [deploy_deps]

        dependencies[service['name']] = set(deploy_deps)

    return dependencies

def generate_bitbucket(config, path):
    bitbucket_config = {
        'image': config['ci']['image'],
        'pipelines': { 'default': [] },
    }

    generate_bitbucket_pipeline(config['ci']['steps'], bitbucket_config['pipelines']['default'])

    if path:
        write_yaml(bitbucket_config, '%s/bitbucket-pipelines.yml' % path)
    else:
        import json
        print(json.dumps(bitbucket_config, indent=2))

def generate_bitbucket_pipeline(steps, pipeline=[]):
    for ci_step in steps:
        step = {}

        if ci_step['type'] == 'parallel':
            step[ci_step['type']] = []
            generate_bitbucket_pipeline(ci_step['steps'], step[ci_step['type']])
        else:
            step[ci_step['type']] = {
                'script': ci_step['commands'],
            }
            if ci_step.get('docker'):
                step[ci_step['type']]['services'] = ['docker']

        pipeline.append(step)


def write_yaml(data, file_path):
    import yaml

    with open(file_path, 'w') as outfile:
        yaml.dump(data, outfile, default_flow_style=False)

    autogen_text = '# AUTOGENERATED FILE (build/deploy.py)\n# python build/deploy.py --generate-ci'
    with open(file_path, 'r') as original: data = original.read()
    with open(file_path, 'w') as modified: modified.write('%s\n\n%s\n%s' % (autogen_text, data, autogen_text))
